#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <hash.hpp>
#include <dpmi.h>

#include "trmsg.h"
#include "trsmt.h"

#define DELIMITER       " \n\t"		//delimiter for parsing config file
#define CONFIGFILE      "TRMGR.CFG"	//config file
#define MAXMESSAGES	10		//maximum number of messages to 
					//process at a time


typedef int (* ActionFn)(char *szLine, int type);

/* data type for hash table */
struct AdapterEntry {
	NameType szAdapter;             //adapter address
	NameType szName;                //adapter name
	unsigned nErrors[10];           //soft error counts
	unsigned nBeaconType;           //beacon type
	AdapterEntry *pNaun[3];         //naun pointers
	AdapterEntry *pNext[3];         //next pointer
	time_t time[3];                 //timestamp
};

static enum OPCODE {FN_SAVE, FN_RESETERROR, FN_RESETFAULT, FN_RESETALL} 
	Opcode;



static int match(AdapterEntry *a, AdapterEntry *b);
static void InorderFunction(AdapterEntry *p);
static unsigned AdapterEntry_hash(AdapterEntry *p, unsigned tablesize);
static const char *xlate(AdapterEntry *p);

static int ValidAddress(char *szAdapter);
static AdapterEntry *ProcessMsg(Msgs * pMsg, const int type);
static void BinToString(BYTE *bin, char *szString);
static int CharToBin(char ch);
static AdapterEntry *CreateEntry(char * szAdapter, char * szName);
static int GetConfig(char *szLine, int type);
static int GetError(char *szLine, int type);
static int GetFault(char *szLine, int type);
static void LinkEntry(AdapterEntry *np, int id);
static void InitConfig(Msgs *request);

static AdapterEntry *pHead[3];
static AdapterEntry *pRingHead;
static FILE *fp;
static ActionFn action[3] = {GetConfig, GetError, GetFault};
static enum Status status = NORMAL;


/* Declare a hash table for type AdapterEntry, with matching function match,
   table size 1024, automatic purging of table contents, and visit
   function InorderFunction. */

declare(zGHSearch,AdapterEntry);
zGHSearch(AdapterEntry) hasher(match,1024,AdapterEntry_hash,1,InorderFunction);


/* We also need to provide functions to compare AdapterEntry objects, a
   visit function to be executed when table entries are visited by
   the inorder function, and a hash number calculation function
   for AdapterEntry objects.  Notice that the match function does
   not have to do anything sophisticated in this case.  It just needs
   to return zero if the objects are equal, nonzero otherwise. */

static int match(AdapterEntry *a, AdapterEntry *b)
{
    return strcmp(a->szAdapter,b->szAdapter);
}

/* functions we want to do when traversing the hash table */
static void InorderFunction(AdapterEntry *p)
{

    switch (Opcode) {
	case FN_SAVE:
	  fprintf(fp, "%s %s\n", p->szAdapter, p->szName);
	  break;
	case FN_RESETERROR:
	  memset(p->nErrors, 0, sizeof(p->nErrors));
	  p->pNaun[SM_ERROR] = NULL;
	  p->pNext[SM_ERROR] = NULL;
	  pHead[SM_ERROR] = NULL;
	  break;
	case FN_RESETFAULT:
	  p->pNaun[SM_FAULT] = NULL;
	  p->pNext[SM_FAULT] = NULL;
	  pHead[SM_FAULT] = NULL;
	  break;
	case FN_RESETALL:
	  memset(p->nErrors, 0, sizeof(AdapterEntry) - sizeof(NameType) * 2);
	  memset(pHead, 0, sizeof(pHead));
	  //pRingHead = NULL;
	  break;
	default:
	  break;
    }
}

/* hashing function */
static unsigned AdapterEntry_hash(AdapterEntry *p, unsigned tablesize)
{
    return zAlphaHash(p->szAdapter, tablesize);
}

/* The reporting function needs a pointer to a function it can use to
   make a string out of an object stored in the table. */
static const char *xlate(AdapterEntry *p)
{
    return (const char *) p->szAdapter;
}


/* check if the token ring network address is a valid one */
static int ValidAddress(char *szAdapter)
{
    int i;

    strupr(szAdapter);
    for (i = 0; i < strlen(szAdapter); i++)
	if ((szAdapter[i] < '0') || (szAdapter[i] > '9'))
	   if ((szAdapter[i] < 'A') || (szAdapter[i] > 'F'))
		return 0;

    return 1;
}


/* create a linked list structure for the ring configuration. Create an
   entry in the database if necessary */
static void InitConfig(Msgs *request)
{
    AdapterEntry entry;

   //get TRMGR's adapter address

   if (pRingHead == NULL)
      {
      BinToString(request->adapter, entry.szAdapter);
      pRingHead = hasher.seek(&entry);
      if (pRingHead == NULL)
	 pRingHead = CreateEntry(entry.szAdapter, entry.szAdapter);
      pHead[SM_CONFIG]  = pRingHead;
      }
}


/* build a linked list of items in the hash table */
static void LinkEntry(AdapterEntry *np, int type)
{
    AdapterEntry *np1;
    
    if (pHead[type] == NULL) 
	{
	pHead[type] = np;
	np->pNext[type] = NULL;
	}
    else
	{
	np1= pHead[type];
	while ((np1 != np) && (np1 != NULL)) np1 = np1->pNext[type];
	if (np1 == NULL)
	   {
	   np->pNext[type] = pHead[type];
	   pHead[type] = np;
	   }
	}
}


/* put the information contained in a message into a linked list */
static AdapterEntry *ProcessMsg(Msgs * pMsg, const int type)
{
    AdapterEntry entry;
    AdapterEntry *np1;
    AdapterEntry *np2;

    BinToString(pMsg->adapter, entry.szAdapter);
    np1 = hasher.seek(&entry);
    if (np1 == NULL)
	np1 = CreateEntry(entry.szAdapter, entry.szAdapter);

    BinToString(pMsg->naun, entry.szAdapter);
    np2 = hasher.seek(&entry);
    if (np2 == NULL)
	np2 = CreateEntry(entry.szAdapter, entry.szAdapter);

    np1->pNaun[type] = np2;
    time(&np1->time[type]); 
    
    LinkEntry(np1, type);
    LinkEntry(np2, type);

    return np1;
}


/* create an entry in the hash table */
static AdapterEntry *CreateEntry(char * szAdapter, char * szName)
{
    AdapterEntry *np;

    np = new AdapterEntry;
    memset(np, 0, sizeof(AdapterEntry));
    strcpy(np->szAdapter, szAdapter);
    strcpy(np->szName, szName);
    hasher.insert(np);

    return np;
}


/* convert a 6-digit network address into a string */
static void BinToString(BYTE *bin, char *szString)
{
    int i;
    static char table[16] = "0123456789ABCDEF";

    for (i = 0; i < 6; i++)
	{
	szString[i * 2] = table[bin[i] >> 4];
	szString[i * 2 + 1] = table[bin[i] & 0x0F];
	}
    szString[12] = 0;

}


/* convert a hex char to binaries format */
static int CharToBin(char ch)
{
	if ((ch >= '0') && (ch <= '9'))
	   return (ch - '0');
	else if ((ch >= 'A') && (ch <= 'F'))
	   return (ch - 'A' + 10);
	else
	   return 0;
}


/* format the the specified entry of the adapter information contained in 
   a linked list stored in the hash table into a string for display */
static int GetConfig(char *szLine, int id)
{
    int i;
    AdapterEntry *np;

    i = 0;
    np = pRingHead;
    if (np)
	do {
	   if (id == i++)
	      {
	      sprintf(szLine, "%s\t%s", np->szName, np->szAdapter);
	      return id;
	      }
	   np = np->pNaun[SM_CONFIG];
	}while ((np != pRingHead) && (np != NULL));

    return -1;
}


/* format the specified entry of error information contained in a linked 
   list stored in a hash table into a string for display */
static int GetError(char *szLine, int id)
{
    int i;
    int j;
    long lTotal;
    AdapterEntry *np;
    AdapterEntry *naun;
    char szBuffer[32];

    i = 0;
    np = pHead[SM_ERROR];
    while (np != NULL)
	{
	if (id == i++)
	   {
	   naun = np->pNaun[SM_ERROR];
	   if (naun == NULL) return -1;
	   lTotal = 0L;
	   for (j = 0; j < sizeof(np->nErrors) / sizeof(int); j++)
		lTotal += np->nErrors[j];

	   sprintf(szLine, "%s\t%s\t%s\t%s", np->szName, naun->szName,
		ltoa(lTotal, szBuffer, 10), np->szAdapter);

	   return id;
	   }
	np = np->pNext[SM_ERROR];
	}

    return -1;
}

/* format the specified entry of fault information contained in a linked 
   list stored in a hash table into a string for display */
static int GetFault(char *szLine, int id)
{
    int i;
    long lTotal;
    AdapterEntry *np;
    AdapterEntry *naun;

    i = 0;
    np = pHead[SM_FAULT];
    while (np != NULL)
	{
	if (id == i++)
	   {
	   naun = np->pNaun[SM_FAULT];
	   if (naun == NULL) return -1;

	   sprintf(szLine, "%s\t%s\t%s\t%s", _ctime(&np->time[SM_FAULT]),
		np->szName, naun->szName, np->szAdapter);


	   return id;
	   }
	np = np->pNext[SM_FAULT];
	}

    return -1;
}


extern "C" {

/* return the number of entries of the requested information in the hash 
   table  */
int GetAdapterInfo(int idInfo, int id, char *szLine)
{
   int i;
   AdapterEntry *np;

   if (idInfo > 2)
	return -1;
   else if (id == SA_GETCOUNT)
	{
	i = 0;
	np = pHead[idInfo];
	while (np != NULL)
	   {
	   i++;
	   np = np->pNext[idInfo];
	   }

	return i;
	}
    else
	return action[idInfo](szLine, id);

    return -1;
}


/* change the symbolic name of the adapter in the hash table */
int ChangeAdapterName(char *szAdapter, char *szName)
{
    AdapterEntry entry;
    AdapterEntry *np;

    if ((strlen(szAdapter) == 0) || (strlen(szName) == 0))
	return -1;

    strncpy(entry.szAdapter, szAdapter, 12);
    entry.szAdapter[12] = 0;
    if ((np = hasher.seek(&entry))== NULL)
	return -1;

    strncpy(np->szName, szName, 12);
    np->szName[12] = 0;

    return 0;
}


/* format the detailed adapter information contained in a linked list stored 
   in a hash table into a string for display */
int GetAdapterDetails(char *szAdapter, AdapterDetails *pDetails)
{
    AdapterEntry entry;
    AdapterEntry *np;
    int i;

    if (strlen(szAdapter) == 0)
	return -1;

    strncpy(entry.szAdapter, szAdapter, 12);
    entry.szAdapter[12] = 0;
    if ((np = hasher.seek(&entry))== NULL)
	return -1;

    strcpy(pDetails->szAdapterName, np->szName);
    strcpy(pDetails->szAdapterAddress, np->szAdapter);
    
    if (np->pNaun[SM_CONFIG])
	{
	strcpy(pDetails->szNaunName, (np->pNaun[SM_CONFIG])->szName);
	strcpy(pDetails->szNaunAddress, (np->pNaun[SM_CONFIG])->szAdapter);
	}
    else    
	{
	strcpy(pDetails->szNaunName, "");
	strcpy(pDetails->szNaunAddress, "");
	}
	
    if (np->pNaun[SM_ERROR])
	{
	strcpy(pDetails->szErrorNaunName, (np->pNaun[SM_ERROR])->szName);
	strcpy(pDetails->szErrorNaunAddress, (np->pNaun[SM_ERROR])->szAdapter);
	}
    else
	{
	strcpy(pDetails->szErrorNaunName, "");
	strcpy(pDetails->szErrorNaunAddress, "");
	}
	
    if (np->pNaun[SM_FAULT])
	{
	strcpy(pDetails->szFaultNaunName, (np->pNaun[SM_FAULT])->szName);
	strcpy(pDetails->szFaultNaunAddress, (np->pNaun[SM_FAULT])->szAdapter);
	}
    else
	{
	strcpy(pDetails->szFaultNaunName, "");
	strcpy(pDetails->szFaultNaunAddress, "");
	}
	
    pDetails->FaultTime = np->time[SM_FAULT];
    pDetails->ErrorTime = np->time[SM_ERROR];
    pDetails->nBeaconType = np->nBeaconType;
    for (i = 0; i < 10; i++)
	pDetails->nErrorCount[i] = np->nErrors[i];

    return 0;
}


/* poll the TRAGN and process the information returned in the messages */
BYTE Poll(Status *pStatus)
{
    AdapterEntry entry;
    AdapterEntry *np;
    Msgs request;
    int i;
    int nCount = MAXMESSAGES;

    do {
	request.msg = MSG_REQUEST_MSG;		
	request.more = 0;
	TRMsg(&request);

	switch (request.msg) {

	   case MSG_DATA_INDICATION:
		   
		if (request.ringstatus & (STS_SIGNALLOSS | STS_WIREFAULT |
			STS_REMOVAL1 | STS_REMOVERECVD))
	       	   status = REMOVED;
		else if (request.ringstatus & STS_XMITBEACON)
	           status = BEACONING;
		else if (request.ringstatus & STS_RECOVERY)
	           status = RECOVERY;
		else
	           status = NORMAL;

		break;

	  case MSG_DATA_ERROR:

		request.msg = MSG_REQUEST_MSG;
		np = ProcessMsg(&request, SM_ERROR);
		for (i = 0; i < 5; i++)
		   np->nErrors[i] += request.isoerr[i];

		for (i = 0; i < 5; i++)
		   np->nErrors[i + 5] += request.noniso[i];

		request.msg = MSG_REQUEST_MSG;
		break;

	  case MSG_DATA_CONFIG:

		request.msg = MSG_REQUEST_MSG;
		np = ProcessMsg(&request, SM_CONFIG);
		request.msg = MSG_REQUEST_MSG;
		break;

	case MSG_DATA_BEACON:

		np = ProcessMsg(&request, SM_FAULT);
		np->nBeaconType = request.beacontype;

		break;


	default:

		break;

	}

   } while ((request.more) && (--nCount));

   *pStatus = status;
   return 0;
}
	       

/* retrieve the information on network address/symbolic name from the file
   TRMGR.CFG */
void ReadConfig(void)
{

    char szLine[128];
    char *szAdapter;
    char *szName;

    if ((fp = fopen(CONFIGFILE, "r")) == NULL)
	return;

    while (fgets(szLine, sizeof(szLine), fp))
	{
	if (((szAdapter = strtok(szLine, DELIMITER)) != NULL) &&
	    (ValidAddress(szAdapter)))
	   {
	   if ((szName = strtok(NULL, DELIMITER)) != NULL)
		CreateEntry(szAdapter, szName);
	   else
		CreateEntry(szAdapter, szAdapter);
	   }
	}

    fclose(fp);

}


/* save the network address/symbolic names in the file TRMGR.CFG */
void SaveConfig(void)
{

    if ((fp = fopen(CONFIGFILE, "w")) == NULL)
	return;

    Opcode = FN_SAVE;
    hasher.inorder();

    fclose(fp);

}


/* reset all or a particular category of information in the hash table */
int ResetInfo(int id)
{
    int i;

    switch (id) {
	case SM_CONFIG:
	   for (i = 0; i < 3; i++)
		pHead[i] = NULL;
	   //pRingHead = NULL;
	   Opcode = FN_RESETALL;
	   hasher.inorder();
	   break;
	case SM_ERROR:
	   pHead[id] = NULL;
	   Opcode = FN_RESETERROR;
	   hasher.inorder();
	   break;
	case SM_FAULT:
	   pHead[id] = NULL;
	   Opcode = FN_RESETFAULT;
	   hasher.inorder();
	   break;

	default:
	   return -1;
    }

    return 0;
}


/* request TRAGN to remove the specified adapter from the ring */
BYTE RemoveIt(char *szAdapter)
{

   int i;
   Msgs msg;

   memset(&msg, 0, sizeof(msg));
   msg.msg = MSG_REQUEST_REMOVE;
   for (i = 0; i < 12; i += 2)
	msg.adapter[i / 2] = CharToBin(szAdapter[i]) * 16 +
               	CharToBin(szAdapter[i + 1]);

   TRMsg(&msg);
   return msg.retcode;

}

/* request TRAGN to reset the TRMGR adapter */
BYTE ResetIt(void)
{
   Msgs request;

   request.msg = MSG_REQUEST_RESET;
   request.more = 0;
   TRMsg(&request);

   if (request.retcode) return request.retcode;
   InitConfig(&request);
   status = NORMAL;

   return 0;
}

/* activate TRAGN to receive MAC frames */
BYTE ActivateIt(BOOL *bEmulate)
{
   Msgs request;

   request.msg = MSG_REQUEST_ACTIVATE;
   request.more = 0;
   TRMsg(&request);

   if (request.retcode) return request.retcode;
   InitConfig(&request);

   *bEmulate = (request.ringstatus == SERVINT);
   return 0;

}


/* inactivate TRAGN so that it stops receiving MAC frames */
BYTE InactivateIt(void)
{
   Msgs request;

   request.msg = MSG_REQUEST_INACTIVATE;
   request.more = 0;
   TRMsg(&request);

   return request.retcode;
}


/* check if TRAGN has been installed */
BOOL TRAGNInstalled(void)
{
   return (dpmi_GetRealModeVector(SERVINT) != 0L);
}


/* routine to remove the \n char at the end of the string containing
   the time */
char * _ctime(const time_t * time)
{
    char *pString;
    char *pIndex;

   pString = ctime(time);

   /* remove the \n char inserted by ctime: bug of ctime? */
   pIndex = strchr(pString, '\n' );
   if (pIndex)
	*pIndex = ' ';

   return pString;
}

}
